#
# Copyright 2008-2014 Universidad Complutense de Madrid
#
# This file is part of Numina
#
# Numina is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Numina is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Numina.  If not, see <http://www.gnu.org/licenses/>.
#

'''Basic tools and classes used to generate recipe modules.

A recipe is a class that complies with the *reduction recipe API*:

 * The class must derive from :class:`numina.core.BaseRecipe`.

'''

import abc
import traceback
import logging

from .. import __version__
from .recipeinout import ErrorRecipeResult
from .recipeinout import RecipeResult as RecipeResultClass
from .recipeinout import RecipeRequirements as RecipeRequirementsClass

_logger = logging.getLogger('numina')


def list_recipes():
    '''List all defined recipes'''
    return BaseRecipe.__subclasses__()


class _BaseRecipeMethods(object):
    '''Base class for all instrument recipes'''

    RecipeResult = RecipeResultClass
    RecipeRequirements = RecipeRequirementsClass

    # Recipe own logger
    logger = _logger

    def __init__(self, *args, **kwds):
        super(_BaseRecipeMethods, self).__init__()
        self.__author__ = 'Unknown'
        self.__version__ = '0.0.0'
        # These two are maintained
        # for the moment
        self.environ = {}
        self.runinfo = {}
        #
        self.instrument = None
        self.configure(**kwds)

    def configure(self, **kwds):
        if 'author' in kwds:
            self.__author__ = kwds['author']
        if 'version' in kwds:
            self.__version__ = kwds['version']
        if 'instrument' in kwds:
            self.instrument = kwds['instrument']
        if 'runinfo' in kwds:
            self.runinfo = kwds['runinfo']

    @classmethod
    def create_result(cls, *args, **kwds):
        '''
        Pass the result arguments to the RecipeResult constructor
        '''
        return cls.RecipeResult(*args, **kwds)

    def run(self, recipe_input):
        return self.RecipeResult()

    def __call__(self, recipe_input):
        '''
        Process the result of the observing block with the
        Recipe.

        :param recipe_input: the input appropriated for the Recipe
        :param type: RecipeRequirement
        :rtype: a RecipeResult object or an error

        '''

        try:
            result = self.run(recipe_input)
        except Exception as exc:
            _logger.error("During recipe execution %s", exc)
            return ErrorRecipeResult(
                exc.__class__.__name__,
                str(exc),
                traceback.format_exc()
                )
        return result

    def set_base_headers(self, hdr):
        '''Set metadata in FITS headers.'''
        hdr['NUMXVER'] = (__version__, 'Numina package version')
        hdr['NUMRNAM'] = (self.__class__.__name__, 'Numina recipe name')
        hdr['NUMRVER'] = (self.__version__, 'Numina recipe version')
        return hdr


class BaseRecipe(_BaseRecipeMethods):
    '''Base class for all instrument recipes'''

    __metaclass__ = abc.ABCMeta

    @abc.abstractmethod
    def run(self, recipe_input):
        return self.RecipeResult()


from .recipeinout import RecipeResult, RecipeRequirements
from .recipeinout import RecipeResultAutoQC
from .dataholders import Product
from .requirements import Requirement


class RecipeType(type):
    '''Metaclass for Recipe.'''
    def __new__(cls, classname, parents, attributes):
        filter_reqs = {}
        filter_prods = {}
        filter_attr = {}

        for name, val in attributes.items():
            if isinstance(val, Requirement):
                filter_reqs[name] = val
            elif isinstance(val, Product):
                filter_prods[name] = val
            else:
                filter_attr[name] = val

        ReqsClass = cls.create_req_class(classname, filter_reqs)

        ResultClass = cls.create_prod_class(classname, filter_prods)

        filter_attr['RecipeResult'] = ResultClass
        filter_attr['RecipeRequirements'] = ReqsClass
        return super(RecipeType, cls).__new__(
            cls, classname, parents, filter_attr)

    @classmethod
    def create_req_class(cls, classname, attributes):
        if attributes:
            reqs_name = '%sRequirements' % classname
            ReqsClass = type(reqs_name, (RecipeRequirements,), attributes)
        else:
            ReqsClass = RecipeRequirements
        return ReqsClass

    @classmethod
    def create_prod_class(cls, classname, attributes):
        if attributes:
            result_name = '%sResult' % classname
            ResultClass = type(result_name, (RecipeResult,), attributes)
        else:
            ResultClass = RecipeResult

        return ResultClass


class RecipeTypeAutoQC(RecipeType):
    '''Metaclass for Recipe with RecipeResultAutoQC.'''
    @classmethod
    def create_prod_class(cls, classname, attributes):
        if attributes:
            result_name = '%sResult' % classname
            ResultClass = type(result_name, (RecipeResultAutoQC,), attributes)
        else:
            ResultClass = RecipeResult

        return ResultClass


class BaseRecipeAlt(_BaseRecipeMethods):
    '''Base class for instrument recipes'''
    __metaclass__ = RecipeType


class BaseRecipeAutoQC(_BaseRecipeMethods):
    '''Base class for instrument recipes'''
    __metaclass__ = RecipeTypeAutoQC
